use std::{
    env::{self, VarError},
    fmt::Debug,
    path::PathBuf,
    str::FromStr,
    time::Duration,
};

pub struct Config {
    pub capture_images: bool,
    pub capture_audio_duration: Option<f64>,
    pub shows_file: PathBuf,
    pub global_tags: Vec<String>,
    pub post_interval: Duration,
    pub cohost_email: String,
    pub cohost_password: String,
    pub cohost_page: String,
    pub cohost_draft: bool,
    pub cohost_cw: bool,
    pub eighteen_plus: bool,
}

const VAR_PREFIX: &str = "SCREENCAP_BOT_";

fn get_var(name: &str) -> Result<String, VarError> {
    env::var(VAR_PREFIX.to_string() + name)
}

fn parse_var<E: Debug, T: FromStr<Err = E>>(name: &str) -> Result<T, VarError> {
    get_var(name).map(|s| {
        s.parse()
            .unwrap_or_else(|e| panic!("Failed to parse {}{}: {:?}", VAR_PREFIX, name, e))
    })
}

fn expect_var(name: &str) -> String {
    get_var(name).unwrap_or_else(|_| panic!("{}{} must be set", VAR_PREFIX, name))
}

pub fn load() -> Config {
    let capture_images = parse_var("CAPTURE_IMAGES").unwrap_or(true);

    Config {
        capture_images,
        capture_audio_duration: parse_var("CAPTURE_AUDIO_DURATION").ok(),
        shows_file: parse_var("SHOWS_FILE").unwrap_or(PathBuf::from("./shows.yaml")),
        global_tags: get_var("GLOBAL_TAGS")
            .map(|s| s.split(',').map(String::from).collect())
            .unwrap_or_default(),
        post_interval: Duration::from_secs(parse_var("POST_INTERVAL").unwrap_or_default()),
        cohost_email: expect_var("COHOST_EMAIL"),
        cohost_password: expect_var("COHOST_PASSWORD"),
        cohost_page: expect_var("COHOST_PAGE"),
        cohost_draft: parse_var("COHOST_DRAFT").unwrap_or(false),
        cohost_cw: parse_var("COHOST_CW").unwrap_or(capture_images),
        eighteen_plus: parse_var("18PLUS").unwrap_or(false),
    }
}
