use std::time::Duration;

use anyhow::{anyhow, Context};
use config::Config;
use lazy_static::lazy_static;
use log::{debug, error, info};
use rand::{
    distributions::{Distribution, Standard, WeightedIndex},
    seq::IteratorRandom,
    Rng,
};
use shows::Shows;

mod config;
mod media;
mod shows;

lazy_static! {
    static ref RETRY_INTERVAL: Duration = Duration::from_secs(30);
}

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    dotenvy::dotenv().ok();
    env_logger::init();
    ffmpeg_next::init().unwrap();

    let mut rng = rand::thread_rng();

    let conf = config::load();

    if let (false, None) = (conf.capture_images, conf.capture_audio_duration) {
        return Err(anyhow!(
            "At least one of image capture and audio capture must be enabled!"
        ));
    }

    info!("Loading shows from {}", conf.shows_file.display());
    let shows = shows::load(&conf.shows_file).with_context(|| {
        format!(
            "Failed to load shows from file {}",
            conf.shows_file.display()
        )
    })?;
    if shows.is_empty() {
        return Err(anyhow!("Shows file is empty!"));
    }

    let dist = WeightedIndex::new(shows.iter().map(|s| s.weight))
        .context("Failed to load show weights")?;

    info!("Logging into cohost as {}", conf.cohost_email);
    let session = eggbug::Session::login(&conf.cohost_email, &conf.cohost_password)
        .await
        .context("Failed to login to cohost")?;

    loop {
        let result = post_random_capture(&conf, &shows, &session, &dist, &mut rng)
            .await
            .context("Failed to post a random capture");

        if conf.post_interval == Duration::ZERO {
            return result;
        }

        let delay = match result {
            Ok(()) => conf.post_interval,
            Err(e) => {
                error!("{}", e);
                *RETRY_INTERVAL
            }
        };
        tokio::time::sleep(delay).await;
    }
}

async fn post_random_capture<R: Rng>(
    conf: &Config,
    shows: &Shows,
    session: &eggbug::Session,
    dist: &WeightedIndex<f32>,
    rng: &mut R,
) -> anyhow::Result<()> {
    let show = &shows[dist.sample(rng)];
    let episodes = show.episodes().with_context(|| {
        format!(
            "Failed to get episode list for show {} with path {}",
            show.title,
            show.path.display()
        )
    })?;
    let (num, file) = episodes.iter().choose(rng).unwrap();

    let descriptor = shows::display_show_episode(show, num);

    info!("Selected: {} - {}", descriptor, file.display());

    let media_info = media::get_media_info(file, Some("eng"))
        .with_context(|| format!("Failed to get info for media file {}", file.display()))?;
    debug!(
        "Media duration: {}",
        format_timestamp(media_info.duration_secs, None)
    );
    debug!(
        "Subtitle stream index: {:?}",
        media_info.subtitle_stream_index
    );

    let max_timestamp = match conf.capture_audio_duration {
        Some(d) => media_info.duration_secs - d,
        None => media_info.duration_secs,
    };
    let timestamp = max_timestamp * rng.sample::<f64, _>(Standard);
    let formatted_timestamp = format_timestamp(timestamp, Some(media_info.duration_secs));
    info!("Taking capture at {}", formatted_timestamp);

    let mut attachments = Vec::new();

    if conf.capture_images {
        let image_data = media::take_screencap(file, timestamp, media_info.subtitle_stream_index)
            .await
            .context("Failed to take screencap")?;

        let image_size = imagesize::blob_size(&image_data)
            .context("Failed to get image size for screencap image data")?;

        let image_attachment = eggbug::Attachment::new(
            image_data,
            format!("{} @{}.png", descriptor, formatted_timestamp),
            String::from("image/png"),
            eggbug::MediaMetadata::Image {
                width: Some(image_size.width as u32),
                height: Some(image_size.height as u32),
            },
        )
        .with_alt_text(format!(
            "Screencap of {} at {}",
            descriptor, formatted_timestamp
        ));

        attachments.push(image_attachment);
    }

    if let Some(duration) = conf.capture_audio_duration {
        let audio_data = media::take_audio_clip(file, timestamp, duration)
            .await
            .context("Failed to take audio clip")?;

        let audio_attachment = eggbug::Attachment::new(
            audio_data,
            format!("{} @{}.mp3", descriptor, formatted_timestamp),
            String::from("audio/mpeg"),
            eggbug::MediaMetadata::Audio {
                artist: show.title.clone(),
                title: descriptor.clone(),
            },
        );

        attachments.push(audio_attachment);
    }

    let mut tags = show.tags.clone();
    tags.extend_from_slice(&conf.global_tags);

    let chost_id = session
        .create_post(
            &conf.cohost_page,
            &mut eggbug::Post {
                content_warnings: if conf.cohost_cw {
                    vec![descriptor]
                } else {
                    vec![]
                },
                attachments,
                tags,
                draft: conf.cohost_draft,
                adult_content: show.eighteen_plus.unwrap_or(conf.eighteen_plus),
                headline: String::new(),
                markdown: String::new(),
                metadata: None,
                ask: None,
            },
        )
        .await
        .context("Failed to create chost")?;
    info!("Created post {}", chost_id);

    Ok(())
}

fn format_timestamp(timestamp: f64, total_duration: Option<f64>) -> String {
    let total_duration = total_duration.unwrap_or(timestamp);
    format!(
        "{}{:02}:{:05.2}",
        if total_duration >= 3600.0 {
            format!("{}:", (timestamp / 3600.0) as u32)
        } else {
            String::new()
        },
        ((timestamp % 3600.0) / 60.0) as u32,
        timestamp % 60.0
    )
}
